/************************************************************************/
/* This file contains the codes for Halton generators in the "bare"	*/
/* form presented in "Fast, portable, and reliable algorithm for ....",	*/
/* and full Halton vector generation, and also an unseccussful		*/
/* "experimental" generator based on binary Halton.			*/
/*									*/
/*			   M. Kolar 1992				*/
/************************************************************************/

/*
List of functions in this file:
""""""""""""""""""""""""""""""
	double init_HaltonBS(base, n0) int base, n0;
	double HaltonBS()
	void init_Halton(base, n0) int base, n0;
	double Halton()
	void init_Halton2(n0) unsigned n0;
	double Halton2()
	void init_vHalton(nbase, n0) int nbase, *n0;
	void rand_init_vHalton(nbase, seed) int nbase; unsigned seed;
	double vHalton()
	void init_ExprG(n0,shft) unsigned n0, shft;
	double ExprG()
*/

#include <errno.h>
#include <time.h>
#include <sys/times.h>
#include "1D.h"


/*-----------------------------------------------------------------------*/
/* Basic algorithms for the Halton numbers as presented
   in the paper "Fast, portable, and reliable algorithm for ...."
   by M. Kolar & S.F. O'Shea. (Computers & Mathematics with Applications)
   First that by "M. Berblinger & C. Schlier, Computer Phys. Commun. 66
   (1991), 157" traslated into C, followed by the new fast algorithms
   by M. Kolar:  */

static int b;
static double Nn, Dn;

double init_HaltonBS(base, n0) int base, n0;
/*========================================*/
{
 /* n0 is the index of the first Halton number to be
    returned by HaltonBS().  */

 double HaltonBS();

 b = base;
 Nn = 0;
 Dn = 1;
 while(--n0 > 0) HaltonBS();
}

double HaltonBS()
/*=============*/
{
 /* On each call, this function returns next Halton number
    if properly initialized by init_HaltonBS().  */

 register double Y;
 register double X;

 X = Dn - Nn;
 if(X == (double)1.) { Nn = 1.;
                       Dn *= b;
                     }
 else   { Y = Dn / b;
          while(X <= Y) Y /= b;
          Nn = (1. + b) * Y - X;
        }
 return Nn / Dn;
}


#define D  22  /* to generate at most b^D Halton numbers */
static unsigned int d[D + 1];
static double r[D + 1];
/* static int b;  already defined above */

void init_Halton(base, n0) int base, n0;
/*====================================*/
{
 /* n0 is the index of the first Halton number to be
    returned by Halton(). */

 unsigned int last, i;

 b = base;
 if(n0 < 1) n0 = 1;
 --n0;

 /* Invert the order of digits of (n0-1) written in the
    base-b notation:  */
 last = 0;
 while(n0 >= b) { d[last++] = n0 % b; n0 /= b; }
 d[last] = n0;
 for(i = last + 1; i < D; i++) d[i] = 0;
 r[D] = 0;
 for(i = D; i > 0; i--) r[i - 1] = (d[i] + r[i]) / b;
}

double Halton()
/*===========*/
{
 register unsigned int l = 0;
 register unsigned int i;

 if(++d[0] == b)
    { do d[l++] = 0; while(++d[l] == b);
      r[l - 1] = (d[l] + r[l]) / b;
      for(i = l - 1; i > 0 ; i--) r[i - 1] = r[i] / b;
      return r[0] / b;
    }
 else return (d[0] + r[0]) / b;
}


#define BITSPERBYTE     8 
#define BITS(type)      (BITSPERBYTE * (int)sizeof(type))
#define HIBITI  (1 << BITS(int) - 1)
#define MAXINT  (~HIBITI)
/* on some systems, the above definitions can be found in a standard
   include file such as <values.h> for UNIX/SUN or Ultrix/DECstation */

static unsigned int HI_BIT = (1 << BITS(int) - 2);
static double MAXHALTp1;  /*   = (MAXINT + 1.);  see below */
                        /*  = 2147483648. for 32 bit integers */
static unsigned int inv_n;  /* to store the inverted order of bits */

void init_Halton2(n0) unsigned n0;
/*==============================*/
{
 int bit;

 /* n0 is the index of the first Halton number to be
    returned by Halton().
    Here base = 2.  */

 MAXHALTp1 = MAXINT + 1.;

 if(n0 < 1) n0 = 1;
 --n0;

 /* To invert the order of bits, first transfer the lowest bit
    from n0 to inv_n. Then repeatedly shift n0 by one bit to
    the right, inv_n by one bit to the left, and each time
    transfer the lowest bit from n0 to inv_n: */
 inv_n = n0 & 01;
 for(bit = 2; bit < BITS(int); bit++)
	 { inv_n <<= 1; inv_n |= ((n0 >>= 1) & 01); }
}

double Halton2()
/*============*/
{
 register unsigned int bit;

 if(inv_n & HI_BIT)
    { inv_n &= ~HI_BIT;
      bit = HI_BIT >> 1;
      while(inv_n & bit) { inv_n &= ~bit; bit >>= 1; }
      inv_n |= bit;
    }
 else inv_n |= HI_BIT;

 return inv_n / MAXHALTp1;
}

/* ----------------------------------------------------------------------- */


/* ------------  full Halton vector using first nBase primes  ------------ */

#define MAXBASE 60
static unsigned int md[MAXBASE][D + 1];
static double mr[MAXBASE][D + 1];
static int base[MAXBASE], iBase, nBase;
int N0[MAXBASE]; /* can be initiated in the calling module */
/* static int b;  already defined above */

void init_vHalton(nbase, n0) int nbase, *n0;
/*========================================*/
{
 /* n0[iBase] is the index of the first Halton number to be
    returned by Halton() for the base base[iBase]. */

 unsigned int last, i;
 extern void Eratosthenes();

 if(nbase > MAXBASE)
	{ PE" MAXBASE = %d in init_vHalton() not large enough!\n", MAXBASE);
	  PE" Must be set larger than or equal to %d.\n\n", nbase);
	}
 base[0] = -1;
 nBase = nbase;
 Eratosthenes(base, nBase);

/*	   PE"base:");
	   for(iBase = 0; iBase < nBase; iBase++) PE" %d", base[iBase]);
	   PE"\nn0:");
	   for(iBase = 0; iBase < nBase; iBase++) PE" %d", N0[iBase]);
	   PE"\n");
*/
 for(iBase = 0; iBase < nBase; iBase++)
 { b = base[iBase];
   if(n0[iBase] < 1) n0[iBase] = 1;
   --n0[iBase];

  /* Invert the order of digits of (n0[iBase]-1) written in the
     base-b notation:  */
   last = 0;
   while(n0[iBase] >= b) { md[iBase][last++] = n0[iBase] % b; n0[iBase] /= b; }
   md[iBase][last] = n0[iBase];
   for(i = last + 1; i < D; i++) md[iBase][i] = 0;
   mr[iBase][D] = 0;
   for(i = D; i > 0; i--) mr[iBase][i - 1] = (md[iBase][i] + mr[iBase][i]) / b;
 }

 iBase = 0;
}

void rand_init_vHalton(nbase, seed) int nbase; unsigned seed;
/*=========================================================*/
{
/*
#ifdef ultrix
 extern void srand();

 srand(seed);
#else
 extern int srand();

 srand((int)seed);
#endif
*/

 for(iBase = 0; iBase < nbase; iBase++)
      N0[iBase] = (rand() / 2147483648.) * 50000;
 init_vHalton(nbase, N0);
}

double vHalton()
/*============*/
{
/* register unsigned int l = 0;   it may actually slow down the calculation!!!
   register unsigned int i;
*/
 unsigned int l = 0;
 unsigned int i, *d;
 double res, *r;

 b = base[iBase];
 d = md[iBase];
 if(++d[0] == b)
    { r = mr[iBase];
      do d[l++] = 0; while(++d[l] == b);
      r[l - 1] = (d[l] + r[l]) / b;
      for(i = l - 1; i > 0 ; i--) r[i - 1] = r[i] / b;
      res = r[0] / b;
    }
 else res = (d[0] + mr[iBase][0]) / b;
 ++iBase;  if(iBase == nBase) iBase = 0;
 return res;
}

/* ----------------------------------------------------------------- */


/*----------------------  experimental generator  ------------------------*/
/* Derived from the code for the generator of the quasi-random Halton
   numbers  halton1R2b()  (see above)
*/

void init_ExprG(n0,shft) unsigned n0, shft;
/*=======================================*/
{
 init_Halton2(n0);
 HI_BIT >>= shft;
}

double ExprG()
/*==========*/
{
 register unsigned int b, nc;
unsigned cBit;
struct tm *localtime();
struct tm *tms;
time_t time();
time_t seconds;

 if(inv_n & HI_BIT)
    { /* at least one carry needed */
      inv_n &= ~HI_BIT;
      b = HI_BIT >> 1;
      while(inv_n & b) { inv_n &= ~b; b >>= 1; }
      inv_n |= b;
    }
 else inv_n |= HI_BIT;

/*	(void)time(&seconds);
	tms = localtime(&seconds);
	cBit = 1 << (int)(tms->tm_sec/2); */

	for(nc = 0; nc < 7; nc++){
	  cBit = 1 << (int)(rand() * 32. / 2147483648.);
          if(inv_n & cBit)  inv_n ^= cBit;  else  inv_n |= cBit;
	}
/*      inv_n |= cBit; */

 return inv_n / MAXHALTp1;
}
/*-------------------  experimental generator  --------------------------*/

