/*****************************************************************************/
/********************** (C) 1992 by Miroslav Kolar ***************************/
/*									     */
/*	Romberg integration of the effective 1D estimators obtained by	     */
/*  the approximate analytic integration (up to the 2nd order in potential)  */
/*  in the N_fict-1 dimensional subspace perpendicular to the body diagonal, */
/*	    of the path-integral quantum Monte-Carlo estimators		     */
/*    For a single particle in a one-dimensional external potential V(x).    */
/*	 This version is for N_fict = 2 fictitious particles only!!!	     */
/*									     */
/*  Internally, potential is assumed to be scaled to certain atomic-like     */
/*  units, so that the hamiltonian is  -(1/2)(d^2/dx^2) + V(x)		     */
/*									     */
/*****************************************************************************/

/*
External files required: 1Dpotential.c
			 NRqromb.c
			 LinePlots.c
			 DateAndTime.c

 List of functions in this file:
	void printChoices(N, descr) int N; char *descr;
	void usage(mode) int mode;
	void defaults(mode) int mode;
	void help1()
	void help()
	main(argc, argv)  int argc; char *argv[];
	void checkCL()
	void dynDefs()
	void printParams()
	void OutDensity()
	void writeOut()
	double fq1(u)  double u;
	double fq2(u)  double u;
	double fal1(u)  double u;
	double fal2(u)  double u;
	double fla1(u)  double u;
	double fla2(u)  double u;
	double fev1(u)  double u;
	double fev2(u)  double u;
	float DensIntegrate(X)  double X;
	float rhoHOexactPrim2(X)  double X;

External functions:
	char *DateAndTime();			(in  DateAndTime.c)
	void initPlot(), plotFun1(), Histogram();
						(in  LinePlots.c)
        Currently available potentials are defined in 1Dpotential.c:
		double (*V)();
		double (*dV)();	     = dV(x)/dx
		double (*epsVir)();  = V(x) + 0.5*(x-Shift)* dV(x)/dx
		void   (*inV)();     input of the parameters of the potential
		long   (*outV)();    output of the parameters of the potential
		double (*V2_4)();  = V''(x) / 4
		double (*V23_2)();  = (V''(x) + 0.25 * (x-Shift) * V'''(x)) / 2
		void choosePotential();

	Romberg integration:
		double qromb(func)  double (*func)();   (in NRqromb.c)
		void qromb_init()			(	"    )
*/

#include <errno.h>
#include <sys/times.h>
#include <stdio.h>                  
#include <math.h>
#include <string.h>

/* #define DEBUG */

#define PE         (void)fprintf(stderr,
#define PR         (void)fprintf(Listing,
#define SC         (void)scanf(
#define FC         (void)fclose(
#define ILLEGAL    { PE" Illegal option '%s' ignored!\n",argv[i]); illeg++; }
#define FIERROR(x) { PR" File '%s': %s\n", x, sys_errlist[errno]); exit(-1); }

#define MAXLETT    101   /* max length of file names, text lines, etc.
			    Must be at least 10  */

#define CUBE	   0  /* Integration inside the square only */
#define ALL	   1  /* Space perpendicular to diagonal extended to infinity */

#define AVERAGES   0
#define RHO1	   1
#define RHOm	   2

/* Defaults:  */
#define DEFpotFlag	0	/* see 1Dpotential.c   */
#define DEFPlotWidth	65
#define DEFN_x		10	/* # of points to calculate density at  */
#define DEFEPSromb	1e-7	/* rel accuracy in Romberg integration  */
#define DEFJMAXromb	18	/* max # of steps in Romberg integration  */
#define DEFKromb	5	/* # of points in Romberg extrapolation  */


/* ------ Global declarations: ------- */
#ifdef ultrix
extern int  errno;
#else
int  errno;
#endif
extern char *sys_errlist[];
extern double exp(), erf(), sqrt();
extern clock_t times();
extern void exit();
/*
#ifdef ultrix
        extern void *malloc();
#else
        extern char *malloc();
#endif
*/

extern char *DateAndTime();			/* in DateAndTime.c  */
extern void initPlot(), plotFun1(), Histogram();
						/* in LinePlots.c    */
extern double qromb();
extern void qromb_init();                       /* in NRqromb.c      */

static struct tms cpuT;		/*    - " -        */

static char *author = "                     Written by M. Kolar (C) 1992\n",
	    *space = "\n                                    ",
	    *progName;	/*  to be set equal to  argv[0]  */

static int outsideFlag = CUBE,
	   actionFlag = AVERAGES; /* : Q and energy averages calculated */
		   /* = RHO1 : rho(xValue) calculated			*/
		   /* = RHOm : whole array of rho(.) calculated		*/
static char *ALLstring = "INFTY",
	    *DENSstring = "rho";

static int  PlotWidth = DEFPlotWidth;
		/*  Width (in characters) of the line-printer plots  */

FILE *Listing = stdout;
static char ListingFileName[MAXLETT];
	 /* To print out parameters, final results, configurations during
	    mixing, etc.    If no file is given, stdout is used.   */

/*  Here  N_fict = 2  */
/*  ================  */
double  beta = -1.,  /* beta = 1/(k*T);
			k - Boltzman constant, T - temperature  */
	m = -1.,     /* mass of the real particle */
	C1,          /* = N_fict * mAt / (2*pi*hAt*hAt*betaAt)      */
	C2,          /* = N_fict * mAt / (2*hAt*hAt*betaAt*betaAt)  */
	C3,          /* = 2 * C2                                    */
	C4;          /* = N_fict / (2*betaAt)			    */

float 	      xMin = 1e+30, xMax = -1e+30; /*  integration limits   */
static double xValue = 0., /* to calculate a single density at (DENS on) */
	      NormFactor;  /* to normalize Q, density */

/* Density of (fictitious) particles: */
static int    N_x = DEFN_x;	  /*  # of points to calculate density at  */
static double *xd;	/*  At xd[i] calculate 		    | dyn defined  */
float	      *dens;	/*  density dens[i]		    | in dynDefs() */

/* Integrals: */
static double
   alfa, lambda, eps_vir,
	  /* See Herman, Bruskin, Berne, J.Chem.Phys. 76, 5150 (1982) */
   Q;  /* canonical sum (coordinate dependent partition function) */

/* Atomic-like values used consistently in the internal representation:     */
static double	betaAt;		/* mAt = 1, hAt = 1 */
double		xMinAt, xMaxAt,
		LAt;		/*  = (xMaxAt - xMinAt)		     */
static double	xStepAt,	/*  = (xMaxAt - xMinAt) / (N_x - 1)  */
		xValueAt;       /* to calculate a single density at  */
double	Cx = 0.,  /* coord. transformation: x = Cx * xAt	   */
	CE = 0.;  /* energy transformation: E = CE * E_atomic,
					    beta = betaAt / CE	   */
	/* For harmonic oscillator: Cx = CE = 1			   */

#define P1237 (float)1.273e29
#define M1237 (float)-1.273e29
static float  p1273 = P1237, clxMin = P1237, 
	      m1273 = M1237, clxMax = M1237;

/* Romberg: */
double EPSromb = DEFEPSromb;
int JMAXromb = DEFJMAXromb,
    Kromb = DEFKromb;

static double SQRTPI; /* = sqrt(Pi) */

/*-----------------------------------------------------------------------*/
/*     Currently available potentials are defined in 1Dpotential.c       */

static int potFlag = DEFpotFlag;
	 /* equals to the number of the selected potential */

double (*V)();		/* pointer to the potential function		*/
double (*dV)();		/* pointer to dV(x)/dx				*/
double (*epsVir)();	/* pointer to the function returning
			V(x) + 0.5*(x-Shift)* dV(x)/dx  for the selected V(x),
			where Shift is nonzero for a potential defined in
			finite interval only, and is selected so that
			the virial estimator gives correct averages	  */
void   (*inV)();	/* pointers to functions for input and output   */
long   (*outV)();	/* of the parameters of the potential		*/
float  xRet1, xRet2;  /* (approximate) classical return points of the
                          potential for energy = Vminimum + 2 / beta,
                          or some other suitable limits			*/
extern double (*V2_4)(),   /* pointer to  V''(x) / 4			 */
	      (*V23_2)(); /* pointer to [V''(x)+0.25*(x-Shift)*V'''(x)]/2  */
extern char *potBlurb,   /* potential description for printParams()      */
	    *potLabel[], /* potential name abbreviations to be used on
			    command line [and in help() and short help]  */
	    *potHelp[];  /* more information on potentials for help()    */
extern int  numPot,      /* total number of currectly defined potentials */
	    HOpotFlag;   /* the value of potFlag for Harmonic Oscillator */
extern void choosePotential();  /* and perform potential initialization  */

/* Unused here but required in 1Dpotential.c: */
FILE *Input;
int N_fict = 2;
void err_fseek(p) char *p; { }
void err_fread(p) char *p; { }

/*	End of definitions for the file 1Dpotential.c			 */
/*-----------------------------------------------------------------------*/

/* Some of the functions defined below the Main:  */
void   checkCL(), dynDefs(), OutDensity(), writeOut();
float  DensIntegrate(), rhoHOexactPrim2();
double fq1(), fq2(), fal1(), fal2(), fla1(), fla2(), fev1(), fev2();


void printChoices(N, descr) int N; char *descr[];
/*=============================================*/
{
int n;

PE"{"); PE *descr);
for(n = 1; n < N; n++) { PE" "); PE descr[n]); }
PE"}");
}

void usage(mode) int mode;
/*======================*/
{
PE"\7\nUsage:"); if(mode) PE"\n");
PE"  %s options [ > file]\n", progName);
if(mode)
 PE"\n  options: pPotential");
 else
 { PE"  options: p");
   printChoices(numPot, potLabel);
 }
PE" lListFile\n");
PE"           bBeta xXmin XXmax hNhist eEPS jJMAX kRextr pwPlotWidth\n");
PE"           {%s %sXvalue %s}                     ( -h  ?  -help )\n",
		ALLstring, DENSstring, DENSstring);
}

void defaults(mode) int mode;
/*=========================*/
{
PE"Defaults:\n ");
PE"  p%s  h%d  e%.16g  j%d  k%d  pw%d\n", potLabel[DEFpotFlag],
		DEFN_x, DEFEPSromb, DEFJMAXromb, DEFKromb, DEFPlotWidth);
PE"     Integration inside the square only\n\n");
if(mode)  PE"Example: %s  b0.4 rho h12 x0.23 X15.0\n\n", progName);
}


void help1()
/*========*/
{
usage(0);
defaults(0);
PE"########### If you need more help, enter command: %s -help ###########\n\n",
	progName);
}

void help()
/*=======*/
{
	int n;
PE"     Romberg integration of the effective 1D estimators obtained by\n");
PE"  the approximate analytic integration (up to the 2nd order in V(x)) in\n");
PE"  the N_fict-1 dimensional subspace perpendicular to the body diagonal,\n");
PE"         of the path-integral quantum Monte-Carlo estimators\n");
PE"   For a single particle in a one-dimensional external potential V(x).\n");
PE"      This version is for N_fict = 2 fictitious particles only.\n");
PE author);
usage(1);
PE"\n  Only some of these options need be given in any order. All essential\n");
PE"  parameters not supplied will be prompted for.\n\n");
PE"  - As a default, energy averages are calculated and the approximate\n");
PE"    analytic integration is limited to a square. \n");
PE"    To see what happens if the integration is extended to infinity\n");
PE"    in the direction perpendicular to the diagonal of the square,\n");
PE"    specify the option %s.\n\n", ALLstring);
PE"  Potential:  "); PE potLabel[0]); PE potHelp[0]);
for(n = 1; n < numPot; n++) {
PE"\n              "); PE potLabel[n]); PE potHelp[n]);
			    }
PE"\n\n  Beta:       1/(kT)\n\n");
PE"  Xmin, Xmax:  integration limits (along one side of the square)\n\n");
PE"  Nhist: # of points to calculate the density at (and nothing else).\n");
PE"         Applies only when  %s  is specified (without any Xvalue)\n\n",
					 DENSstring);
PE"  Xvalue: only a single density calculated at x=Xvalue\n\n");
PE"        Density calculation doesn't require any further numerical\n");
PE"        integration. Analytical integration limited to the square was\n");
PE"        done after higher than 2nd order terms in V(x) were neglected.\n\n");
PE"  EPS: Relative accuracy of the Romberg integration\n\n");
PE"  JMAX: Max # of steps of the Romberg integration\n\n");
PE"  Rextr: # of points used in Romberg extrapolation\n\n");
PE"  PlotWidth: Width (in characters) of all line-printer type plots\n");
PE"             (PlotWidth = 0 means no plotting at all)\n\n");
defaults(1);
exit(0);
}
/* End of help() */


/*=====================================*/
/*    MAIN program starts here!!!!     */
/*=====================================*/
main(argc, argv)  int argc; char *argv[];
/*=====================================*/
{
  int	i, atoi(), n,
	illeg = 1, /* see #define ILLEGAL  */
	clpotFlag;
  long atol();
  int  strcmp(), strncmp();
  char *ARG, *strcpy();
  double atof(), aux;

#ifdef DEBUG
 PE"\nAvailable potentials:  ");
 for(n = 0; n < numPot; n++) { PE potLabel[n]); PE"   "); }
 PE"\n");
#endif

progName = argv[0];
for (i=1; i<argc; i++)
  {
   ARG = argv[i];
#ifdef DEBUG
   PE"\n..... processing '%s'\n",ARG);
#endif
   if ( ! strcmp(ARG,"-help") || ! strcmp(ARG,"-HELP") )  help();
   if ( ! strcmp(ARG,"-h") || ! strcmp(ARG,"-H") ||
	! strcmp(ARG,"?") || ! strcmp(ARG,"-?") )  { help1(); exit(0); }

     else if(! strcmp(ARG,ALLstring))	    outsideFlag = ALL;
     else if(! strncmp(ARG,DENSstring,(n = (int)strlen(DENSstring))) )
           { if((int)strlen(ARG) == n) actionFlag = RHOm; else
		{ actionFlag = RHO1; ARG += n; xValue = atof(ARG); }
	   }
     else if(*ARG == 'e')
           { aux = atof(++ARG); if(aux > 0.) EPSromb = aux; else ILLEGAL }
     else if(! strncmp(ARG,"pw",2))
       { ARG += 2; if(isdigit(*ARG)) PlotWidth = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'p')
           { ++ARG; for(n = 0; n < numPot; n++)
			{ if(! strncmp(ARG,potLabel[n],strlen(potLabel[n])) )
						{ clpotFlag = n; break; }
			}
             if( clpotFlag == -1) { ILLEGAL } else potFlag = clpotFlag;
           }
     else if(*ARG == 'l') strcpy(ListingFileName,++ARG);
     else if(*ARG == 'h')
       { ++ARG; if(isdigit(*ARG)) N_x = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'j')
       { ++ARG; if(isdigit(*ARG)) JMAXromb = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'k')
       { ++ARG; if(isdigit(*ARG)) Kromb = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'x') clxMin = (float)atof(++ARG);
     else if(*ARG == 'X') clxMax = (float)atof(++ARG);
     else if(*ARG == 'b')
           { aux = atof(++ARG); if(aux > 0.) beta = aux; else ILLEGAL }
     else ILLEGAL
  }

if (illeg == argc) help1(); 

/* Initialize Listing File:  */
if(ListingFileName[0] != '\0')
 { if( (Listing = fopen(ListingFileName,"A")) == NULL)
				  FIERROR(ListingFileName)
 }

PR space); PR DateAndTime());
PR"\n\n    %s:  ", progName);
if(actionFlag == AVERAGES) {
 PR"Romberg integration of the effective 1D estimators obtained by\n");
 PR"   the approximate 2nd-order N_fict-1 dimensional analytic integration\n");
 PR"         of the path-integral quantum Monte-Carlo estimators\n");
 PR author); PR"\n");
}
else {
 PR"Approximate density obtained by analytical integration after\n");
 PR"    higher than 2nd-order terms in V(x) were neglected in the primitive\n");
 PR"            path-integral quantum Monte-Carlo density matrix\n");
}
 PR"             For a single particle in a V(x). N_fict = 2.\n");
 PR author); PR"\n");

checkCL();
dynDefs();
SQRTPI = sqrt(M_PI);

switch (actionFlag)
 {
   case AVERAGES: if(outsideFlag == CUBE)
#ifdef DEBUG
			{ PE"Q:\n");
			  Q = qromb(fq1);
			  PE"alfa:\n");
			  alfa = qromb(fal1);
			  PE"lambda:\n");
			  lambda = qromb(fla1);
			  PE"eps_vir:\n");
			  eps_vir = qromb(fev1);
			}
			else
			{ PE"Q:\n");
			  Q = qromb(fq2);
			  PE"alfa:\n");
			  alfa = qromb(fal2);
			  PE"lambda:\n");
			  lambda = qromb(fla2);
			  PE"eps_vir:\n");
			  eps_vir = qromb(fev2);
			}
#else
			{ Q = qromb(fq1);
			  alfa = qromb(fal1);
			  lambda = qromb(fla1);
			  eps_vir = qromb(fev1);
			}
			else
			{ Q = qromb(fq2);
			  alfa = qromb(fal2);
			  lambda = qromb(fla2);
			  eps_vir = qromb(fev2);
			}
#endif
		  writeOut();
		  break;
   case	RHO1:	  PR"\n rho(%g) = %g\n\n", xValue, DensIntegrate(xValueAt));
		  if(potFlag == HOpotFlag)
			PR"\tExact primitive value for N_fict=2 is %g\n\n",
						rhoHOexactPrim2(xValueAt));
		  break;
   case	RHOm:	  for(i = 0; i < N_x; i++) dens[i] = DensIntegrate(xd[i]);
		  OutDensity();
 }

(void)times(&cpuT);
PR"Total (us+sys) CPU time used by the whole job: %g sec\n\n",
	(float)(cpuT.tms_utime + cpuT.tms_stime) / 60.);
FC Listing);
exit(0);

}  /* ------------------------ end of Main --------------------------  */


void checkCL()
/*============         Use all command line values that make sense  */
{
void printParams();

if(clxMin == p1273 && clxMax == m1273) PE"\n  CHECK xMin AND xMax !!!\n\n");
if(clxMin != p1273) xMin = clxMin;
if(clxMax != m1273) xMax = clxMax;

while(beta <= 0.0)
	 { PE"\7Enter positive value of beta: "); SC"%lf", &beta); }

choosePotential(potFlag, 0);  /* among others, it calculates
		      Cx and CE, and C1, C2 and C3 */

while(xMax <= xMin)
	{ PR"\nFor integration limits, enter xMin < xMax: ");
	  SC"%f%f", &xMin, &xMax);
	}

if( actionFlag == RHO1 && (xValue < xMin || xValue > xMax) )
	{
	 PR"\n x = %g to calculate the density at is outside of (%g,%g)!\n\n",
			xValue, xMin, xMax);
	 exit(-1);
	}

xMaxAt = xMax / Cx;
xMinAt = xMin / Cx;
LAt = xMaxAt - xMinAt;
xStepAt = LAt / (N_x - 1);
betaAt = beta * CE;
C4 = 1. / betaAt;  /* N_fict = 2 */
if(actionFlag == RHO1) xValueAt = xValue / Cx;

/* Calculate normalization factor: */
NormFactor = C1;
 
/* Print out all parameters, plot potential: */
printParams();

}  /* ------------------- end of checkCL() -----------------------  */


void dynDefs()
/*==========*/
{
 int  i;

 if(actionFlag == AVERAGES) qromb_init();
 if(actionFlag == RHOm)
   {  xd = (double *)malloc(N_x * sizeof(double));
      dens = (float *)malloc(N_x * sizeof(float));
      xd[N_x - 1] = xMaxAt;
      for(i = N_x - 2; i >= 0; i--) xd[i] = xd[i + 1] - xStepAt;
   }
}


void printParams()
/*==============*/
/* Print out all important parameters at the beginning of the listing.  */
/* (Called at the end of  checkCL() )					*/
{
 static char *fmt = "%8ld ", *fmtV = "%9g";
 static int  fmt_len = 9;
 int i;

 PR potBlurb); (void)(*outV)(Listing);
 PR"Beta = %.8g\n", beta);
 PR"No. of fictitious particles = 2\n");
 PR"xMin = %g,  xMax = %g\n", xMin, xMax);
 if(actionFlag == AVERAGES)
	   { PR"Analytic integration ");
	     if(outsideFlag == CUBE) PR"limited to the square\n");
		else  PR"extended to infinity\n");
	     PR"Romberg: EPS = %g, JMAX = %d, K = %d\n",
					EPSromb, JMAXromb, Kromb);
           }
	else
	   PR"Analytic integration limited to the square\n");

 if(PlotWidth)
   { /* Initialize line-printer type plotting routines:  */
     initPlot((float)Cx, (float)CE, (float)xMinAt, (float)xMaxAt,
			1, PlotWidth, fmt, fmt_len, Listing);

     /* Plot potential:    */
     PR"\n");
     for(i = -2 * fmt_len; i < PlotWidth; i += 2) PR" ");
     PR"V(x)\n");
     plotFun1(V, fmtV);
   }
 PR"\n");
}


void OutDensity()
/*=============*/
{
int i;
float maxD = 0;

PR"\n\t\t\tDensity (normalized to Q):\n");
if(PlotWidth)
  { for(i = 0; i < N_x; i++)
	if(fabs((double)dens[i]) > maxD) maxD = dens[i];

    if(maxD > 1000.) PR"\n\t\t     Density is too big to be plotted!\n");
    else Histogram(dens, N_x, (float)(xMinAt - 0.5 * xStepAt), (float)xStepAt);
  }

PR"\n    x           rho(x)\n");
for(i = 0; i < N_x; i++) PR"  %g\t    %g\n", Cx * xd[i], dens[i]);
if(potFlag == HOpotFlag)
  { PR"\n    x    Exact primitive rho(x) for N_fict=2\n");
    for(i = 0; i < N_x; i++)
	PR"  %g\t    %g\n", (maxD = Cx * xd[i]), rhoHOexactPrim2(maxD));
  }
PR"\n");
}  /* --------------->  end of:  OutDensity()  <------------------ */


void writeOut()
/*===========*/
{
   double eps;

   PR"Q = %.16g\n", Q * NormFactor * M_SQRT2);
   alfa *= ( C3 / Q );
   lambda /= Q;
   eps = CE * (C4 - alfa + lambda);
   eps_vir *= (CE / Q);
   PR"<alfa> = %.16g, <lambda> = %.16g\n", CE * alfa, CE * lambda);
   PR"<eps> = %.16g, <eps_vir> = %.16g\n", eps, eps_vir);
   if(potFlag == HOpotFlag)
    { double b8 = 8. + betaAt * betaAt,
	     b16 = 8. + b8,
	     B = CE / (betaAt * b16);

      PR"\tExact values for N_fict=2: \tQ = %.16g\n", 4./(betaAt * sqrt(b16)));
      PR"\t<alfa> = %.16g, <lambda> = %.16g\n", 8.* B, (lambda = b8 * B));
      PR"\t<E> = <eps_vir> = %.16g\n", 2. * lambda);
    }
   PR"\n");
   (void)fflush(Listing);
}


#define ETA(u)  (M_SQRT2 * ((u1 = (u) - xMinAt) < (u2 = xMaxAt - (u)) ? u1 : u2))
#define ALFA(u) (betaAt * (4. * C2 + (V2 > 0. ? V2 : 0.)))

double fq1(u)  double u;
/*====================*/
{
double	u1, u2,
	V2 = (*V2_4)(u),
	sqAL = sqrt(ALFA(u));

return exp(-betaAt * (*V)(u)) * SQRTPI * erf(sqAL * ETA(u)) / sqAL;
}

double fq2(u)  double u;
/*====================*/
{
double V2 = (*V2_4)(u);

return exp(-betaAt * (*V)(u)) * sqrt(M_PI / ALFA(u));
}

double fal1(u)  double u;
/*====================*/
{
double	u1, u2,
	V2 = (*V2_4)(u),
	AL = ALFA(u),
	ET = ETA(u),
	sqAL = sqrt(AL);

return exp(-betaAt * (*V)(u)) *
	(SQRTPI * erf(sqAL * ET) / sqAL - 2. * ET * exp(-AL * ET * ET)) / AL;
}

double fal2(u)  double u;
/*====================*/
{
double V2 = (*V2_4)(u), AL = ALFA(u);

return exp(-betaAt * (*V)(u)) * sqrt(M_PI / AL) / AL;
}

double fla1(u)  double u;
/*====================*/
{
double	u1,
	u2 = (*V2_4)(u),
        V2 = (u2 > 0. ? u2 : 0.),
	AL = betaAt * (4. * C2 + V2),
	ET = ETA(u),
	sqAL = sqrt(AL),
	v = (*V)(u),
	B = SQRTPI * erf(sqAL * ET) / sqAL;

return exp(-betaAt * v) *
	(v * B + V2 * (B * 0.5 - ET * exp(-AL * ET * ET)) / AL);
}

double fla2(u)  double u;
/*====================*/
{
double	u2 = (*V2_4)(u),
        V2 = (u2 > 0. ? u2 : 0.),
	AL = betaAt * (4. * C2 + V2),
	v = (*V)(u);

return exp(-betaAt * v) *
	(v + 0.5 * V2 / AL) * sqrt(M_PI / AL);
}

double fev1(u)  double u;
/*====================*/
{
double	u1,
	u2,
	V2 = (*V2_4)(u),
	AL = ALFA(u),
	ET = ETA(u),
	sqAL = sqrt(AL),
	B = SQRTPI * erf(sqAL * ET) / sqAL;

return exp(-betaAt * (*V)(u)) *
    ((*epsVir)(u) * B + (*V23_2)(u) *
			 (B * 0.5 - ET * exp(-AL * ET * ET)) / AL);
/*
u1 = (*V23_2)(u);
return exp(-betaAt * (*V)(u)) *
    ((*epsVir)(u) * B + (u1 > 0. ? u1 : 0.) *
			 (B * 0.5 - ET * exp(-AL * ET * ET)) / AL);
*/
}

double fev2(u)  double u;
/*====================*/
{
double	V2 = (*V2_4)(u), AL = ALFA(u);

return exp(-betaAt * (*V)(u)) *
	((*epsVir)(u) + 0.5 * (*V23_2)(u) / AL) * sqrt(M_PI / AL);
}
/*
{
double	u1, V2 = (*V2_4)(u), AL = ALFA(u);

u1 = (*V23_2)(u);
return exp(-betaAt * (*V)(u)) *
	((*epsVir)(u) + 0.5 * (u1 > 0. ? u1 : 0.) / AL) * sqrt(M_PI / AL);
}
*/

float DensIntegrate(X)  double X;
/*=============================*/
/* Returns densities scaled back by 1/Cx to external units */
{
 double a, b, aBeta, xU, xL, v, aux = (*V2_4)(X);

 v = (*V)(X);
 a = C3 + (aux > 0. ? aux : 0.);
 b = 0.5 * (*dV)(X);
 if(a == 0.)
	{ if(b == 0.) return (float)(NormFactor * exp(-betaAt * v) * LAt / Cx);
	  aux = b * 0.5;
	  return (float)( NormFactor *
			  ( exp(betaAt * (aux * (X - xMinAt) - v))
			    - exp(betaAt * (aux * (X - xMaxAt) - v)) )
			  / (aux * betaAt * Cx)
			);
	}
 xL = 0.5 * (aux = b / a) - X;
 xU = xL + xMaxAt;
 xL += xMinAt;
 aBeta = sqrt(a * betaAt);
 return  (float)( NormFactor * exp(betaAt * (0.25 * b * aux - v))
		  * fabs( erf(aBeta * fabs(xU))
		      + (xL * xU > 0 ? -1. : 1.) * erf(aBeta * fabs(xL)) )
		  / (M_2_SQRTPI * aBeta * Cx)
		);
				/*   here M_2_SQRTPI = 2 / sqrt(pi)   */

}  /* --------------->  end of:  DensIntegrate()  <------------------ */


float rhoHOexactPrim2(X)  double X;
/*===============================*/
{
 double b = 8. + betaAt * betaAt;

 return (float)( 2. * exp(-0.25 * betaAt * (8. + b) * X * X / b)
		 / (sqrt(M_PI *betaAt * b) * Cx) );
}
